/* Copyright (C) 2011 Papavasileiou Dimitris                             
 *                                                                      
 * This program is free software: you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation, either version 3 of the License, or    
 * (at your option) any later version.                                  
 *                                                                      
 * This program is distributed in the hope that it will be useful,      
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        
 * GNU General Public License for more details.                         
 *                                                                      
 * You should have received a copy of the GNU General Public License    
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <limits.h>
#include <math.h>

#include "state.h"
#include "config.h"

static double samples[3][SAMPLING_WINDOW_SIZE];
static double kernel[SAMPLING_WINDOW_SIZE];
static unsigned int initialized;
static char blink, subblink;

static unsigned char indices[3], fill[3], update;
static unsigned long overflows = 10;

void update_engine_state(void)
{
    double sum;
    int i;

    if (!initialized) {
	double sigma, sum;

	sigma = (double) SAMPLING_WINDOW_SIZE / 3;

	/* Calculate a gaussian smoothing
	   kernel of width SAMPLING_WINDOW_SIZE */

	for (i = 0, sum = 0 ; i < SAMPLING_WINDOW_SIZE ; sum += kernel[i], i += 1) {
	    kernel[i] =
		(1 + (i > 0)) * exp (-i * i / (2 * sigma * sigma)) / (sqrt (2 * M_PI) * sigma);
	}

	for (i = 0 ; i < SAMPLING_WINDOW_SIZE ; i += 1) {
	    kernel[i] /= sum;
	}

	initialized = 1;
    }

    if (update) {
	if (fill[0] == SAMPLING_WINDOW_SIZE) {
	    /* Update engine rpm. */

	    for (i = 0, sum = 0;
		 i < SAMPLING_WINDOW_SIZE;
		 sum += samples[0][i] * kernel[i], i += 1);

	    engine.speed = F_CPU / 8 / PULSES_PER_REVOLUTION * 60 / sum;
	}
	
	update = 0;
    } else if (overflows >= 10) {
	/* When overflows keep piling up it means the
	   motor has stopped turning. */
    
	overflows = 0;
	engine.speed = 0;
    }
    
    /* Sample coolant temperature and battery voltage. */
    
    if (timing.subblink != subblink) {
	double V, R, R_2, R_3, R_4;
	
	unsigned int sample;
	unsigned char l, h;

	/* Power up the sensor. */
	
	COOLANT_DDR |= (1 << COOLANT_BIT);
	COOLANT_PORT |= (1 << COOLANT_BIT);
	
	/* Select the proper ADC input channel and
	   AVcc as reference.*/

	ADMUX = (1 << REFS0) | COOLANT_CHANNEL;

	/* Wait for the differential channel and the sensor to
	   stabilize and start a conversion. */
	
	_delay_us(200);
	ADCSRA |= (1 << ADSC); 

	while (ADCSRA & (1 << ADSC));

	/* Power down the sensor. */

	COOLANT_PORT &= ~(1 << COOLANT_BIT);
	    
	/* Take a reading and add it to the sliding window. */
	
	l = ADCL;
	h = ADCH;

	sample = (unsigned short)l | ((unsigned short)h << 8);

	if (sample & (1 << 9)) {
	    /* The sample is negative so sign-extend it
	       and take the absolute value. */
	    
	    sample = abs((int)(sample | 0xfc00));
	}

	/* Calculate the sampled voltage in Volts and
	   the sesnor's resistance in Ohms. */
	
	V = sample / 512.0 / COOLANT_GAIN * COOLANT_VREF;
	R = -COOLANT_RESISTOR * V / (V - COOLANT_VREF);

	if (R < COOLANT_THRESHOLD) {
	    /* Convert the measured sensor resistance into a
	       temperature and add it to the sliding window.
	       This uses a 4-th degree polynomial function
	       which can be fround from measured temperature-resistance
	       pairs using Octave's polyfit for example. */
	    
	    R_2 = R * R;
	    R_3 = R_2 * R;
	    R_4 = R_2 * R_2;

	    samples[1][indices[1]] =
		COOLANT_COEFF_4 * R_4 + COOLANT_COEFF_3 * R_3 +
		COOLANT_COEFF_2 * R_2 + COOLANT_COEFF_1 * R +
		COOLANT_COEFF_0;
	} else {
	    samples[1][indices[1]] = -1;
	}
	
	indices[1] = (indices[1] + 1) % SAMPLING_WINDOW_SIZE;

	if (fill[1] < SAMPLING_WINDOW_SIZE) {
	    fill[1] += 1;
	}
	
	/* Select the proper ADC input channel for
	   battery voltage and set AVcc as reference.*/

	ADMUX = (1 << REFS0) | VOLTAGE_CHANNEL;
	_delay_us(200);
	ADCSRA |= (1 << ADSC);
	while (ADCSRA & (1 << ADSC));

	/* Take a reading and add it to the sliding window.
	   Engine voltage is assumed to be read through a
	   resistor divider to bring it down to the range
	   [0, VOLTAGE_VREF]. */
	
	l = ADCL;
	h = ADCH;

	samples[2][indices[2]] = ((unsigned short)l | ((unsigned short)h << 8)) * (VOLTAGE_R1 + VOLTAGE_R2) / VOLTAGE_R2 * VOLTAGE_VREF / 1023;
	indices[2] = (indices[2] + 1) % SAMPLING_WINDOW_SIZE;

	if (fill[2] < SAMPLING_WINDOW_SIZE) {
	    fill[2] += 1;
	}
	
	subblink = timing.subblink;
    }
    
    if (timing.blink != blink) {
	if (fill[1] == SAMPLING_WINDOW_SIZE) {
	    /* Update the coolant temperature reading. */

	    for (i = 0, sum = 0;
		 i < SAMPLING_WINDOW_SIZE;
		 sum += samples[1][i] * kernel[i], i += 1);

	    temperature.coolant = sum;
	}
    
	/* Update the battery reading. */

	if (fill[2] == SAMPLING_WINDOW_SIZE) {
	    for (i = 0, sum = 0;
		 i < SAMPLING_WINDOW_SIZE;
		 sum += samples[2][i] * kernel[i], i += 1);

	    engine.voltage = sum;
	}
	
	blink = timing.blink;
    }
}
    
ISR(TIMER1_OVF_vect)
{
    overflows += 1;
}

ISR(TIMER1_CAPT_vect)
{
    static unsigned long previous, olddelta;
    unsigned long newdelta, sample;

    /* If the current timestamp is "in the past" assume an
       overflow has occured so add 2^16.  This is not accurate
       for deltas larger than 2^16 but these correspond to
       engine speeds way outside the range of interest. */
    
    if (ICR1 > previous) {
	newdelta = (unsigned long)ICR1 - previous;
    } else {
	newdelta = (unsigned long)ICR1 + 65536 - previous;
    }

    previous = ICR1;
    overflows = 0;

#ifdef CBR_400_RR_NC29
#define THRESHOLD 5000
    /* Look ahead to see what the future holds.  We need this
       so that we may know what pulse pattern to expect. This is
       due to my particular ECU behaving erratically (through
       fault or design).  In the case where the tacho signal
       always sends N pulses per revolution sample should always
       be equal to newdelta (see below). */
    
    if (!olddelta) {
	olddelta = newdelta;
    } else {
	if ((long)newdelta - (long)olddelta > THRESHOLD) {
	    sample = newdelta + olddelta;
	    olddelta = 0;
	} else {
	    sample = olddelta;
	    olddelta = newdelta;
	}

	/* Put the calculated sample into the window. */
	
	samples[0][indices[0]] = sample;
	indices[0] = (indices[0] + 1) % SAMPLING_WINDOW_SIZE;

	if (fill[0] < SAMPLING_WINDOW_SIZE) {
	    fill[0] += 1;
	}

	update = 1;
    }
#else
    /* Put the calculated sample into the window. */
	
    samples[0][indices[0]] = newdelta;
    indices[0] = (indices[0] + 1) % SAMPLING_WINDOW_SIZE;

    if (fill[0] < SAMPLING_WINDOW_SIZE) {
	fill[0] += 1;
    }

    update = 1;
#endif
}

